@@ -1,16 +1,16 @@
-/* ZGV v2.0 - (c) 1993,1994 Russell Marks for improbabledesigns.
- * See README for license details.
+/* Zgv v2.3 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
+ * Copyright (C) 1993, 1994 Russell Marks. See README for license details.
  *
  * helppage.c - for the online help page displays.
  */
 
 #include <stdio.h>
 #include <string.h>
+#include <unistd.h>
 #include <vga.h>
 #include "3deffects.h"
 #include "font.h"
 #include "readnbkey.h"
-#include "handlevt.h"
 #include "rc_config.h"
 #include "rcfile.h"
 
@@ -18,8 +18,6 @@
 
 helpscreenon()
 {
-/* aha!!! Finally diked this bastard out */
-/* vga_setmode(TEXT); */
 vga_setmode(G640x480x16);
 vga_setcolor(15);
 
@@ -60,18 +58,10 @@ while(!quit)
     if(key!=0)
       {
       gotkey=1;
-      /* if it's from F1 to F8, we switch consoles */
-      if((key>=RK_F1)&&(key<=RK_F8))
-        {
-        if(!handlevt(ttyfd,key-RK_F1+1))    /* screen gets redrawn later */
-          gotkey=0;    
-        }
-      else
-        quit=1;
+      quit=1;
       }
       
     usleep(10000);
     }
   }
-/* vga_setmode(TEXT);    hmm... should this still be here? */
 }
