@@ -1,5 +1,5 @@
-/* ZGV v2.0 - (c) 1993,1994 Russell Marks for improbabledesigns.
- * See README for license details.
+/* Zgv v2.3 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
+ * Copyright (C) 1993, 1994 Russell Marks. See README for license details.
  *
  * rc_config.h - Defined! the zgv_config structure.
  *                       photographic evidence on page 2. :)
@@ -13,6 +13,10 @@ struct zgv_config {
   int nodelprompt,perfectindex;
   int xvpic_index;
   int onefile_progress,cleartext;
+  int force16fs;
+  int repeat_timer;
+  int revert,gamma,selecting;
+  int dontchangestdin;
   
   int brightness;
   double contrast;
@@ -21,4 +25,6 @@ struct zgv_config {
   int medium_r,medium_g,medium_b;
   int light_r,light_g,light_b;
   int mode_allowed[256];
+  
+  char cmd[1024];	/* alternative command to run instead of viewing */
   };
