@@ -1,5 +1,5 @@
-/* ZGV v2.0 - (c) 1993,1994 Russell Marks for improbabledesigns.
- * See README for license details.
+/* Zgv v2.3 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
+ * Copyright (C) 1993, 1994 Russell Marks. See README for license details.
  *
  * readnbkey.c - I'd call this inkey$.c if it wasn't so awkward; reads a
  *               non-blocking key. Header file provides #defines for weirdo
@@ -24,7 +24,7 @@ char keybuf[1024];
 
 /* this gets all the characters sent by the key into an ASCIIZ string */
 f=0;
-while(keybuf[f++]=getnbkey(ttyfd));
+while((keybuf[f++]=getnbkey(ttyfd)));
 
 realkey=0;
 /* key=getnbkey(ttyfd);    surely we don't need this! */
