@@ -1,9 +1,11 @@
-/* ZGV v2.0 - (c) 1993,1994 Russell Marks for improbabledesigns.
- * See README for license details.
+/* Zgv v2.3 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
+ * Copyright (C) 1993, 1994 Russell Marks. See README for license details.
  *
  * texterrors.c - provides a fake 'msgbox' routine which printf()'s instead.
  */
 
+#include <stdio.h>
+
 
 int msgbox(ttyfd,message,replytype,lite,dark,txt)
 int ttyfd;
