@@ -1,5 +1,5 @@
-/* ZGV v2.2 - (c) 1993,1994 Russell Marks for improbabledesigns.
- * See README for license details.
+/* Zgv v2.3 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
+ * Copyright (C) 1993, 1994 Russell Marks. See README for license details.
  *
  * vgadisp.c - vga specific display routines.
  */
@@ -20,7 +20,6 @@
 #include "usejpeg.h"
 #include "readpnm.h"
 #include "readnbkey.h"
-#include "handlevt.h"
 #include "helppage.h"
 #include "3deffects.h"
 #include "rc_config.h"
@@ -51,9 +50,6 @@ char viewerhelp[][80]={
   "m \t\t\t mirror",
   "f \t\t\t flip",
   "c \t\t\t toggle centering picture on screen",
-#ifdef HANDS_ON_VC_CODE
-  "F1 to F8 \t\t change virtual console",
-#endif
   "Esc or x \t\t exit the viewer",
   " ",
   "hjkl or qaop keys scroll around the picture",
@@ -80,15 +76,21 @@ byte *image_palette;
 
 double contrast=1.0;  /* note that double contrast is in fact 2 :) */
 int brightness=0;
-int scrnwide,scrnhigh,realscrnwide,scrnpixelsize,scrncols;
+int scrnwide,scrnhigh,scrnpixelsize,scrncols;
+int scaling=1,interp=0,inextpix=1;
 
 int palr[256],palg[256],palb[256];
 byte palr64[256],palg64[256],palb64[256];
 int palrgb[768];
 
 int loading_file_type;
+int repeat_sig,first_repeat;
+int saved_px,saved_py;		/* for persistance in cfg.repeat_timer mode */
 
-int ttyfd;
+extern int zgv_ttyfd;
+
+
+void repeat_sighandler();
 
 
 #define GET15BITCOLOUR(r,g,b) (((r>>3)<<10)|((g>>3)<<5)|(b>>3))
@@ -107,83 +109,94 @@ char *giffn;
 hffunc howfarfunc;
 int show_dont_tell;
 {
-int result;
+int result,sgres;
 PICINFO ginfo;
 byte *palette;
 
 
-ginfo.numcols=256;	/* only changes for GIF files */
+first_repeat=1;
 
-switch(loading_file_type=magic_ident(giffn))
+do
   {
-  case _IS_GIF:
-    /* part of the New Strategy (tm); always use 8-bit for GIFs.
-     * fix the mode to a similar 8-bit one if possible.
-     */
-    if(show_dont_tell) fix_to_similar_mode(FIX_TO_EIGHT_BIT);
-    pixelsize=1;
-    result=readgif(giffn,&theimage,&palette,howfarfunc);
+  ginfo.numcols=256;	/* only changes for GIF files */
+  
+  switch(loading_file_type=magic_ident(giffn))
+    {
+    case _IS_GIF:
+      /* part of the New Strategy (tm); always use 8-bit for GIFs.
+       * fix the mode to a similar 8-bit one if possible.
+       */
+      if(show_dont_tell) fix_to_similar_mode(FIX_TO_EIGHT_BIT);
+      pixelsize=1;
+      result=readgif(giffn,&theimage,&palette,howfarfunc);
+      if(result==_PIC_OK)
+        {
+        getgifinfo(&ginfo);
+        height=ginfo.height; width=ginfo.width;
+        }
+      break;
+  
+    case _IS_JPEG:
+      /* don't *always* use 24-bit for JPEGs... check for 15/16/24-bit
+       * video modes before doing so, as well as the config.
+       */
+      if(show_dont_tell)
+        {
+        pixelsize=has_highcolour_mode()?3:1;
+        if(pixelsize==3 && cfg.jpeg24bit==0)
+          pixelsize=1;
+        }
+  
+      theimage=NULL;
+      result=read_JPEG_file(giffn,howfarfunc,&palette);
+      
+      /* if error, then palette has already been nuked */
+      if(theimage==NULL)
+        result=_PICERR_NOMEM;
+      else
+        if(result!=_PIC_OK) free(theimage);
+        
+      if(result!=_PIC_OK) result=_PICERR_SHOWN_ALREADY;
+      break;
+      
+    case _IS_PNM:
+      if(show_dont_tell) pixelsize=has_highcolour_mode()?3:1;
+      result=read_pnm_file(giffn,howfarfunc,&theimage,&palette,&pixelsize,
+      		&ginfo);
+      width=ginfo.width;
+      height=ginfo.height;
+      if(result!=_PIC_OK)
+        {
+        if(theimage!=NULL) free(theimage);
+        if(palette!=NULL)  free(palette);
+        }
+      break;
+    
+    /* if they voted for "none of the above"... */
+    default:
+      return(_PICERR_BADMAGIC);
+    }
+  
+  
+  if(show_dont_tell)
+    {
+    if(pixelsize==3)
+      fix_to_similar_mode(FIX_TO_HIGHCOLOUR);
+    else
+      fix_to_similar_mode(FIX_TO_EIGHT_BIT);
+  
     if(result==_PIC_OK)
       {
-      getgifinfo(&ginfo);
-      height=ginfo.height; width=ginfo.width;
+      numcols=ginfo.numcols;
+      sgres=showgif(palette);
+      free(theimage);
+      free(palette);
       }
-    break;
-
-  case _IS_JPEG:
-    /* don't *always* use 24-bit for JPEGs... check for 15/16/24-bit
-     * video modes before doing so, as well as the config.
-     */
-    if(show_dont_tell)
-      {
-      pixelsize=has_highcolour_mode()?3:1;
-      if(pixelsize==3 && cfg.jpeg24bit==0)
-        pixelsize=1;
-      }
-
-    theimage=NULL;
-    result=read_JPEG_file(giffn,howfarfunc,&palette);
-    
-    /* if error, then palette has already been nuked */
-    if(theimage==NULL)
-      result=_PICERR_NOMEM;
-    else
-      if(result!=_PIC_OK) free(theimage);
-      
-    if(result!=_PIC_OK) result=_PICERR_SHOWN_ALREADY;
-    break;
-    
-  case _IS_PNM:
-    if(show_dont_tell) pixelsize=has_highcolour_mode()?3:1;
-    result=read_pnm_file(giffn,howfarfunc,&theimage,&palette,&pixelsize,
-    		&ginfo);
-    width=ginfo.width;
-    height=ginfo.height;
-    if(result!=_PIC_OK)
-      {
-      if(theimage!=NULL) free(theimage);
-      if(palette!=NULL)  free(palette);
-      }
-  }
-
-
-if(show_dont_tell)
-  {
-  if(pixelsize==3)
-    fix_to_similar_mode(FIX_TO_HIGHCOLOUR);
-  else
-    fix_to_similar_mode(FIX_TO_EIGHT_BIT);
-
-  if(result==_PIC_OK)
-    {
-    numcols=ginfo.numcols;
-    showgif(palette);
-    free(theimage);
-    free(palette);
     }
+  else
+    image_palette=palette;
   }
-else
-  image_palette=palette;
+while(show_dont_tell && cfg.repeat_timer && result==_PIC_OK && sgres==0);
 
 return(result);
 }
@@ -235,15 +248,11 @@ newmode=-1; pixdiff=(1<<30);
 numpixels=get_mode_width(curvgamode)*get_mode_height(curvgamode);
 
 /* we check 0 - 255 at the most */
-/* Harm's put something in to deal with this, must check that sometime */
-#ifdef HARM_HANEMAAYER_FIXES_THAT_SVGALIB_BUG
 for(f=0;f<256;f++)
-#else
-for(f=0;f<31;f++)
-#endif
   {
   if((!vga_hasmode(f))||(cfg.mode_allowed[f]==0)) continue;
   vminfo=vga_getmodeinfo(f);
+  if(vminfo==NULL) continue;
   
   mode_ok=0;
   if(modetype==FIX_TO_EIGHT_BIT)
@@ -302,12 +311,43 @@ return(isjpeg);
 makerealpal()
 {
 int f;
+int red=299, green=587, blue=114;	/* NTSC uses these */
+int r,rlo,g,glo,b,blo;
+int gval;
 
 for(f=0;f<256;f++)
   {
-  palrgb[f*3  ]=palr64[f]=palr[f]>>2;
-  palrgb[f*3+1]=palg64[f]=palg[f]>>2;
-  palrgb[f*3+2]=palb64[f]=palb[f]>>2;
+  r=palr[f]>>2; rlo=palr[f]&3;
+  g=palg[f]>>2; glo=palg[f]&3;
+  b=palb[f]>>2; blo=palb[f]&3;
+
+  if(cfg.gamma)
+    {
+    gval=(rlo*red+glo*green+blo*blue)/4;
+    /* so gval is sub-value, 0<=gval<1000. (really <750... :( )
+     * possible emulated sub-values are...
+     * 0, 114 (b), 299 (r), 413 (r+b), 587 (g), 701 (g+b), 886 (r+g)
+     */
+  
+    if(gval>=886)
+      r++,g++;
+    else if(gval>=701)
+      g++,b++;
+    else if(gval>=587)
+      g++;
+    else if(gval>=413)
+      r++,b++;
+    else if(gval>=299)
+      r++;
+    else if(gval>=114)
+      b++;
+
+    if(r>63) r=63; if(g>63) g=63; if(b>63) b=63;
+    }
+
+  palrgb[f*3  ]=palr64[f]=r;
+  palrgb[f*3+1]=palg64[f]=g;
+  palrgb[f*3+2]=palb64[f]=b;
   }
 }
 
@@ -421,7 +461,7 @@ else
 if(pixelsize!=1) gl_setcontextvga(curvgamode);
 
 scrnwide=get_mode_width(curvgamode);
-scrnhigh=realscrnwide=get_mode_height(curvgamode);
+scrnhigh=get_mode_height(curvgamode);
 scrnpixelsize=get_mode_pixelbytes(curvgamode);
 scrncols=get_mode_numcols(curvgamode);
 
@@ -444,14 +484,17 @@ graphicsoff()
 }
 
 
-showgif(palette)   /* global theimage,height,width implied */
+/* returns 1 if ESC or 'x' was pressed, 0 if because of timeout when
+ * using cfg.repeat_timer option.
+ */
+int showgif(palette)   /* global theimage,height,width implied */
 byte *palette;
 {
 int opx,opy,px,py,quitshow,key,redraw;
+int npx,npy;    /* equivalent to px,py in scaling>1 mode */
 
 /* if image has less than 256 colours, we fill in the 64 greys so that
  * using the vkludge on, say, a mono file will look much better.
- * mind you, it doesn't appear to work very well...!
  */
 if(numcols<256)
   {
@@ -460,32 +503,48 @@ if(numcols<256)
     palette[f]=palette[256+f]=palette[512+f]=((f-numcols)<<2);
   }
 
-px=py=quitshow=0;
+repeat_sig=0;
+if(cfg.repeat_timer)
+  {
+  if(cfg.repeat_timer==-1)
+    repeat_sig=2;	/* make it look (nearly) instant */
+  else
+    {
+    signal(SIGALRM,repeat_sighandler);
+    alarm(cfg.repeat_timer);
+    }
+  }
+
+quitshow=0;
 filterpal(palette);
-graphicson();
-if((ttyfd=open("/dev/tty",O_RDONLY|O_NONBLOCK))<0)
-  {graphicsoff();
-  printf("Couldn't get nonblocking read access to /dev/tty.\n");
-  exit(1);}  
+if(!cfg.repeat_timer || (cfg.repeat_timer && first_repeat))
+  {
+  px=py=0;
+  graphicson();
+  first_repeat=0;
+  if(cfg.revert)
+    {
+    scaling=1;
+    interp=0;
+    }
+  }
+else
+  if(cfg.repeat_timer)
+    {
+    px=saved_px;
+    py=saved_py;
+    }
+
 redrawgif(px,py);
 
-while(!quitshow)
+while(!quitshow && repeat_sig!=1)
   {
   usleep(10000);
-  key=readnbkey(ttyfd);
+  key=readnbkey(zgv_ttyfd);
   opx=px; opy=py; redraw=0;
 
-#if HANDS_ON_VC_CODE
-  /* if it's from F1 to F8, we switch consoles */
-  if((key>=RK_F1)&&(key<=RK_F8))
-    if(handlevt(ttyfd,key-RK_F1+1))
-      {
-      graphicson();
-      redraw=1;
-      }
-#else
   /* svga modes - note that F11=shift-F1 and F12=shift-F2 */
-  if(((key>=RK_F1)&&(key<=RK_F12))||(key>=RK_SHIFT_F3)&&(key<=RK_SHIFT_F10))
+  if(((key>=RK_F1)&&(key<=RK_F12))||((key>=RK_SHIFT_F3)&&(key<=RK_SHIFT_F10)))
     /* the RK_F?'s are defined such that this sort of thing is ok to do: */
     /* (here comes the 'if' statement from hell...) */
     if((vga_hasmode(key-RK_F1+10))&&(cfg.mode_allowed[key-RK_F1+10])&&
@@ -497,8 +556,8 @@ while(!quitshow)
       graphicson();
       redraw=1;
       }
-#endif
-  if((key>='0')&&(key<='9'))
+
+  if(key=='0' || (key>='5' && key<='8'))
     {
     if(key=='0')
       {
@@ -514,6 +573,7 @@ while(!quitshow)
         {
         curvgamode=key-48;
         if((curvgamode==7)||(curvgamode==8)) virtual=1; else virtual=0;
+        if(virtual) scaling=1;
         graphicson();
         redraw=1;
         }
@@ -522,6 +582,56 @@ while(!quitshow)
   else
     switch(key)
       {
+      case 's': case 'd':
+        if(virtual) break;
+        zoom=0;
+        if(scaling<512)
+          {
+          int oldscale=scaling;
+          
+          scaling+=(key=='d')?scaling:1;
+          if(scaling>512) scaling=512;
+          samecentre(&px,&py,scaling,px,py,oldscale);
+          redraw=1; /* no cls reqd. - must be bigger */
+          }
+        break;
+      case 'S': case 'D':
+        if(virtual) break;
+        zoom=0;
+        if(scaling>1)
+          {
+          int oldscale=scaling;
+          
+          scaling-=(key=='D')?scaling/2:1;
+          if(scaling<1) scaling=1;
+          samecentre(&px,&py,scaling,px,py,oldscale);
+          redraw=1;
+          if(width*scaling<scrnwide || height*scaling<scrnhigh) graphicson();
+          }
+        break;
+      case 'n':
+        if(zoom || scaling>1)
+          {
+          samecentre(&px,&py,1,px,py,scaling);
+          scaling=1; zoom=0;
+          redraw=1; graphicson(); 
+          }
+        scaling=1; redraw=1;
+        if(scaling>1)  { scaling--; redraw=1; graphicson(); }
+        if(scaling==1) { px=npx; py=npy; }
+        break;
+      case RK_HOME: case 'A'-0x40:
+        px=py=0; break;
+      case RK_END:  case 'E'-0x40:
+        px=py=1<<30; break;
+      case RK_PAGE_UP: case 'U'-0x40:
+        py-=scrnhigh*9/10; break;
+      case RK_PAGE_DOWN: case 'V'-0x40:
+        py+=scrnhigh*9/10; break;
+      case '-':
+        px-=scrnwide*9/10; break;
+      case '=':
+        px+=scrnwide*9/10; break;
       case 'v': vkludge=((vkludge==1)?0:1); redraw=1; graphicson(); break;
       case ',': contrast-=0.05;
         filterpal(palette); vga_setpalvec(0,256,palrgb); break;
@@ -544,21 +654,20 @@ while(!quitshow)
       case 'm': fx_mirror(); px=py=0; redraw=1; break;
       case 'f': fx_flip();   px=py=0; redraw=1; break;
       case 'r': fx_rot();    px=py=0; redraw=1; break;
+      case 'R': fx_rot(); fx_flip(); fx_mirror(); px=py=0; redraw=1; break;
       case 'z':
-        zoom=(!zoom); redraw=1; px=py=0; graphicson(); break;
+        zoom=(!zoom); scaling=redraw=1; px=py=0; graphicson(); break;
+      case 'i':
+        interp=(!interp); redraw=1; break;
+      case '1':
+        inextpix=(inextpix==1)?2:1;
+        redraw=1;
+        break;
       case 'c':
         cfg.centreflag=(!cfg.centreflag);
         redraw=1; px=py=0; graphicson(); break;
-#if 0
-/* this appears to be a bit broken with VC switching after ^Z */
-      case 26:   /* ^Z */
-        vga_setmode(TEXT);
-        raise(SIGTSTP);
-        graphicson(); redraw=1;
-        break;
-#endif
       case '?':
-        showhelp(ttyfd,"- KEYS FOR FILE DISPLAY SCREEN -",viewerhelp);
+        showhelp(zgv_ttyfd,"- KEYS FOR FILE DISPLAY SCREEN -",viewerhelp);
         /* falls through to 'refresh screen' */
       case 12: case 18:     /* 12,18 = Ctrl-L, Ctrl-R */
         graphicson(); redraw=1; break;
@@ -567,59 +676,86 @@ while(!quitshow)
       }
   if(!zoom)
     {
-    if(height<=scrnhigh)
+    int swidth=width*scaling,sheight=height*scaling;
+    
+    if(sheight<=scrnhigh)
       py=0;
     else
-      if(height-py<scrnhigh) py=height-scrnhigh;
-    if(width<=scrnwide)
+      if(sheight-py<scrnhigh) py=sheight-scrnhigh;
+    if(swidth<=scrnwide)
       px=0;
     else
-      if(width-px<scrnwide) px=width-scrnwide;
+      if(swidth-px<scrnwide) px=swidth-scrnwide;
     if(px<0) px=0;
     if(py<0) py=0;
+    if(scaling>1)
+      {
+      npx=px/scaling;
+      npy=py/scaling;
+      }
     }
   else
-    px=py=0;
-  if((redraw)||(opx!=px)||(opy!=py)) redrawgif(px,py);
+    px=py=npx=npy=0;
+    
+  if((redraw)||(opx!=px)||(opy!=py)) redrawgif(px,py,npx,npy);
+  if(repeat_sig==2) repeat_sig--;
   }
 graphicsoff();
-close(ttyfd);
+
+if(cfg.selecting)
+  show_dimensions(px,py,scaling);
+
+if(cfg.repeat_timer)
+  {
+  alarm(0);
+  saved_px=px;
+  saved_py=py;
+  }
+  
+return(quitshow);
 }
 
 
-redrawgif(px,py)
-int px,py;
+void repeat_sighandler()
+{
+repeat_sig=1;
+}
+
+
+/* this routine is getting ridiculous */
+redrawgif(px,py,npx,npy)
+int px,py,npx,npy;
 {
 int x,y,xdim;
 int x_add,y_add;
 byte *realline,*ptr;
-long place;
 
 if(zoom)
   drawzoomedgif(height,width,theimage);
 else
   {
+  int swidth=width*scaling,sheight=height*scaling;
+  
   /* draw non-zoomed pic */
   x_add=y_add=0;	/* these control the centering */
   if(cfg.centreflag)
     {
-    if(width<scrnwide)
-      x_add=(scrnwide-width)>>1;
+    if(swidth<scrnwide)
+      x_add=(scrnwide-swidth)>>1;
     
-    if(height<scrnhigh)
-      y_add=(scrnhigh-height)>>1;
+    if(sheight<scrnhigh)
+      y_add=(scrnhigh-sheight)>>1;
       
     if(virtual) x_add>>=1;
     }
   
-  
-  if(width-px<scrnwide) xdim=width-px; else xdim=scrnwide;
-  if((py>=height)||(px>=width)) return(0);
+  if(swidth-px<scrnwide) xdim=swidth-px; else xdim=scrnwide;
+  if((py>=sheight)||(px>=swidth)) return(0);
   /* hopefully the following is fairly quick as fewer ifs... ? */
   if((virtual)&&(pixelsize==1))
     {
     if((realline=calloc(1,scrnwide))==NULL) return(0);
-    for(y=0;(y<height-py)&&(y<scrnhigh);y++)
+    for(y=0;(y<sheight-py)&&(y<scrnhigh);y++)
       {
       for(x=0;x<xdim;x++)
         *(realline+(x>>1))=getvpix(px,py,x,y);
@@ -627,6 +763,116 @@ else
       }
     free(realline);
     }
+  else if(scaling>1)
+    {
+    /* Better grab your Joo Janta 200 Super-Chromatic Peril Sensitive
+     * Sunglasses (both pairs) for this next bit...
+     */
+    int cdown=-1,c1,c2,c3,c4,i,pxx,pyy,pyym,ifres,ifres2,ifres3;
+    int a1,a2,a3,a4,in_rg,in_dn,in_dr;
+    int wp=width*pixelsize,sci=scaling*inextpix;
+    int scaleincr,subpix_xpos,subpix_ypos,sxmulsiz,symulsiz,simulsiz;
+    int sisize,sis2;
+    unsigned char *ptr1,*ptr2,*ptr3,*ptr4;
+    unsigned char *src,*dst,*nextline;
+    
+    /* this doesn't allow for 'virtual' yet! */
+    if((realline=calloc(pixelsize,scrnwide))==NULL) return(0);
+    if(pixelsize==3)
+    if((nextline=calloc(pixelsize,scrnwide))==NULL) return(0);
+    
+    if(interp)
+      {
+      sisize=0;
+      while(sisize<256) sisize+=scaling;
+      scaleincr=sisize/scaling;
+      simulsiz=scaleincr*sisize;
+      sis2=sisize*sisize;
+      }
+    
+    for(y=0,pyy=py;(y<sheight-py)&&(y<scrnhigh);y++,pyy++)
+      {
+      /* this is horribly slow... :( */
+      if(cdown<=0 || interp)
+        {
+        src=theimage+pixelsize*(pyy/scaling)*width;
+        dst=realline;
+        if(pixelsize==1)
+          for(x=0;(x<swidth-px)&&(x<scrnwide);x++)
+            *dst++=*(src+(px+x)/scaling);
+        else if(interp==0)
+          /* normal */
+          for(x=0;(x<swidth-px)&&(x<scrnwide);x++)
+            {
+            ptr=src+((px+x)/scaling)*pixelsize;
+            *dst++=*ptr++; *dst++=*ptr++; *dst++=*ptr;
+            }
+        else
+          {
+          /* interpolated */
+          
+          /* This has been hacked into unreadability in an attempt to get it
+           * as fast as possible.
+           * It's still really slow. :(
+           */
+
+          in_rg=inextpix*3;
+          in_dn=inextpix*wp;
+          in_dr=in_dn+in_rg;
+          pyym=pyy%scaling;
+          ifres2=(pyy>=sheight-sci);
+          subpix_ypos=(pyy%scaling)*scaleincr;
+          subpix_xpos=(px%scaling)*scaleincr;  /* yes px not pxx */
+          
+          ptr1=ptr3=src+(px/scaling)*3;
+          ptr2=ptr4=ptr1+in_rg;
+          if(pyy<sheight-sci)
+            {
+            ptr3=ptr1+in_dn;
+            ptr4=ptr1+in_dr;
+            }
+          
+          symulsiz=sisize*subpix_ypos;
+          sxmulsiz=sisize*subpix_xpos;
+          
+          for(x=0,pxx=px;x<swidth-px && x<scrnwide;x++,pxx++)
+            {
+            a3=symulsiz-(a4=subpix_xpos*subpix_ypos);
+            a2=sxmulsiz-a4;
+            a1=sis2-sxmulsiz-symulsiz+a4;
+
+            for(i=0;i<3;i++)
+              *dst++=(ptr1[i]*a1+ptr2[i]*a2+
+                      ptr3[i]*a3+ptr4[i]*a4)/sis2;
+              
+            subpix_xpos+=scaleincr;
+            sxmulsiz+=simulsiz;
+            if(subpix_xpos>=sisize)
+              {
+              subpix_xpos=sxmulsiz=0;
+              ptr1+=3; ptr2+=3; ptr3+=3; ptr4+=3;
+              }
+
+            if(pxx>=swidth-sci)
+              {
+              ptr2-=3;
+              ptr4-=3;
+              }
+            }
+          }
+              
+        cdown=(cdown==-1)?(scaling-(py%scaling)):scaling;
+        }
+      
+      cdown--;
+      
+      if(scrnpixelsize!=3)
+        eventuallydrawscansegment(realline,x_add,y+y_add,scrnwide*pixelsize);
+      else
+        gl_putbox(x_add,y+y_add,scrnwide,1,realline);
+      }
+    free(realline);
+    }
   else
     {
     switch(scrnpixelsize)
@@ -676,7 +922,7 @@ if(scrnpixelsize!=2)
   vga_drawscansegment(ptr,x,y,len);
 else
   {
-  int f,a;
+  int f;
   unsigned short *bufref;
 
   if(scrncols==32768)
@@ -718,8 +964,7 @@ register int a,b,x,yp,yw;
 long y,sw,sh,lastyp;
 int xoff,yoff;
 int x_add,y_add;	/* for centering */
-int c,cr,cg,cb,pixwide,pixhigh,pr,pg,pb;
-long place;
+int c,cr,cg,cb,pixwide,pixhigh;
 int bigimage;
 byte *rline;
 int tmp1,tmp2,tr,tg,tb,tn;
@@ -869,51 +1114,111 @@ else
 
 fx_mirror()
 {
+unsigned char *src,*dst;
 byte *tmp;
 int x,y;
 
-tmp=malloc(width*height*pixelsize);
+tmp=malloc(width*pixelsize);
 if(tmp==NULL) return(0);
-for(y=0;y<height;y++)
-  for(x=0;x<width;x++)
-    memcpy(tmp+pixelsize*(y*width+(width-x-1)),
-    	theimage+pixelsize*(y*width+x),pixelsize);
 
-free(theimage);
-theimage=tmp;
+if(pixelsize==1)
+  {
+  for(y=0;y<height;y++)
+    {
+    dst=tmp;
+    src=theimage+y*width+width-1;
+    for(x=0;x<width;x++)
+      *dst++=*src--;
+    memcpy(src+1,tmp,width);
+    }
+  }
+else
+  {
+  /* we know pixelsize must be 3 then, so take advantage of that */
+  
+  for(y=0;y<height;y++)
+    {
+    dst=tmp;
+    src=theimage+3*(y*width+width-1);
+    for(x=0;x<width;x++)
+      {
+      *dst++=*src++;
+      *dst++=*src++;
+      *dst++=*src;
+      src-=5;
+      }
+    memcpy(src+3,tmp,width*3);
+    }
+  }
+
+free(tmp);
 }
 
 
 fx_flip()
 {
-byte *tmp;
-int x,y;
+unsigned char *tmp,*ptr1,*ptr2;
+int y,y2,wp=width*pixelsize;
 
-tmp=malloc(width*height*pixelsize);
+tmp=malloc(wp);
 if(tmp==NULL) return(0);
-for(y=0;y<height;y++)
-  memcpy(tmp+pixelsize*((height-y-1)*width),
-  	theimage+pixelsize*(y*width),width*pixelsize);
 
-free(theimage);
-theimage=tmp;
+ptr1=theimage;
+ptr2=theimage+(height-1)*wp;
+
+for(y=0,y2=height-1;y<height/2;y++,y2--,ptr1+=wp,ptr2-=wp)
+  {
+  memcpy(tmp,ptr1,wp);
+  memcpy(ptr1,ptr2,wp);
+  memcpy(ptr2,tmp,wp);
+  }
+
+free(tmp);
 }
 
 
 fx_rot()
 {
-byte *tmp;
-int x,y;
+unsigned char *tmp,*tmp2,*ptr1,*ptr2;
+int x,y,y2,wp=width*pixelsize,hp=height*pixelsize;
+int blockup=4;
 
-tmp=malloc(width*height*pixelsize);
-if(tmp==NULL) return(0);
-for(y=0;y<height;y++)
-  for(x=0;x<width;x++)
-    memcpy(tmp+pixelsize*((height-y-1)+x*height),
-    	theimage+pixelsize*(y*width+x),pixelsize);
+tmp=malloc(wp*height);
+if(pixelsize==3) tmp2=malloc(wp*blockup);
+if(tmp==NULL || (pixelsize==3 && tmp2==NULL)) return(0);
+
+if(pixelsize==1)
+  for(y=0,y2=hp-1;y<height;y++,y2--)
+    {
+    ptr1=tmp+y2; ptr2=theimage+wp*y;
+    for(x=0;x<width;x++,ptr1+=hp)
+      *ptr1=*ptr2++;
+    }
+else
+  {
+  int hpm3=hp-3;
+  
+  for(y=0,y2=height-1;y<height;y++,y2--)
+    {
+    ptr1=tmp+y2*3;
+    if(y%blockup==0)
+      {
+      memcpy(tmp2,theimage+wp*y,wp*((y2<blockup)?y2:blockup));
+      ptr2=tmp2;
+      }
+    for(x=0;x<width;x++,ptr1+=hpm3)
+      {
+      *ptr1++=*ptr2++;
+      *ptr1++=*ptr2++;
+      *ptr1++=*ptr2++;
+      }
+    }
+  }
+  
 x=height;y=width;
 width=x;height=y;
 
+if(pixelsize==3) free(tmp2);
 free(theimage);
 theimage=tmp;
 graphicson(); /* clear screen 'cos image diff. size */
@@ -953,3 +1258,47 @@ for(pr=palr64,pg=palg64,pb=palb64,f=0;f<checknumcols;f++,pr++,pg++,pb++)
 vkcache[idx].rgb=rgb;
 return((int)(vkcache[idx].col=distnum));
 }
+
+
+samecentre(ppx,ppy,newscale,oldpx,oldpy,oldscale)
+int *ppx,*ppy,newscale,oldpx,oldpy,oldscale;
+{
+int xa,ya,sw,sh;
+
+/* even if the centre flag is off, we still need to do this */
+xa=ya=0;
+sw=oldscale*width;
+sh=oldscale*height;
+if(sw<scrnwide) xa=(scrnwide-sw)>>1;
+if(sh<scrnhigh) ya=(scrnhigh-sh)>>1;  
+if(virtual) xa>>=1;
+
+/* finds centre of old screen, and makes it centre of new one */
+*ppx=(oldpx-xa+(scrnwide>>1))*newscale/oldscale;
+*ppy=(oldpy-ya+(scrnhigh>>1))*newscale/oldscale;
+
+xa=ya=0;
+sw=newscale*width;
+sh=newscale*height;
+if(sw<scrnwide) xa=(scrnwide-sw)>>1;
+if(sh<scrnhigh) ya=(scrnhigh-sh)>>1;  
+if(virtual) xa>>=1;
+
+*ppx-=(scrnwide>>1)+xa;
+*ppy-=(scrnhigh>>1)+ya;
+}
+
+
+/* this gets run on picture exit when cfg.selecting is true.
+ * the output is designed to be put straight into a pnmcut command-line.
+ */
+show_dimensions(px,py,scaling)
+int px,py,scaling;
+{
+int x,y,w,h;
+
+x=px/scaling; y=py/scaling;
+if(width*scaling<scrnwide)  w=width;  else w=scrnwide/scaling;
+if(height*scaling<scrnhigh) h=height; else h=scrnhigh/scaling;
+printf("%4d %4d %4d %4d\n",x,y,w,h);
+}
