@@ -1,8 +1,22 @@
-/* ZGV v2.2 - (c) 1993,1994 Russell Marks for improbabledesigns.
- * See README for license details.
+/* Zgv v2.3 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
+ * Copyright (C) 1993, 1994 Russell Marks.
  *
  * zgv.c - This provides the zgv file selector, and interfaces to the
  *         vga display routines (vgadisp.c)
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or (at
+ * your option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
 
@@ -16,6 +30,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/file.h>
+#include <sys/vt.h>
 #include <vga.h>
 
 #include "zgv.h"
@@ -24,7 +39,6 @@
 #include "readnbkey.h"
 #include "font.h"
 #include "3deffects.h"
-#include "handlevt.h"
 #include "helppage.h"
 #include "rc_config.h"
 #include "rcfile.h"
@@ -102,6 +116,11 @@ int zgv_ttyfd,howfar_upto;
 jmp_buf setjmpbuf;   /* in case someone aborts decompressing a file */
 static int xv332_how_far_xpos,xv332_how_far_ypos;
 static int one_file_only=0;
+int original_vt,separate_vt=0;
+
+
+void switchback();
+
 
 
 main(argc,argv)
@@ -110,6 +129,18 @@ char **argv;
 {
 int argsleft;
 
+
+/* kludgetastic, mate */
+if(argc>1 && strcmp(argv[1],"-d")==0) cfg.dontchangestdin=1;
+
+if(fixvt()==0)
+  {
+  fprintf(stderr,"Not running on console and no free VTs found.\n");
+  exit(1);
+  }
+
+atexit(switchback);		/* this gets called after svgalib stuff */
+
 vga_disabledriverreport();    /* quieten svgalib */
 vga_init();
 /* root permissions should now have been ditched */
@@ -117,31 +148,49 @@ vga_init();
 pixelsize=1;
 getconfig();				/* see rcfile.c... */
 argsleft=parsecommandline(argc,argv);	/* ...for these two. */
-if(argsleft>1) usage_help();
+if(argsleft>1) {oldvttell(); usage_help();}
 copyfromconfig();
 
 cleartext();
 
 /* do one-file-only if have arg. */
+/* (does a chdir and carries on if it's a dir.) */
 if(argsleft==1)
   load_one_file(argv[argc-1]);
 
 screenon();
-if((zgv_ttyfd=open("/dev/tty",O_RDONLY|O_NONBLOCK))>=0)
-  {
-  mainloop();
-  close(zgv_ttyfd);
-  }
+openstdin_nonblocking();
+mainloop();
 screenoff();
+
 exit(0);
 }
 
 
+openstdin_nonblocking()
+{
+zgv_ttyfd=fileno(stdin);
+fcntl(zgv_ttyfd,F_SETFL,O_NONBLOCK);
+}
+
+
 load_one_file(filename)
 char *filename;
 {
 hffunc hf;
+struct stat sbuf;
+
+/* before we load it, see if it's a directory. If so, we run zgv with
+ * much the same result as typing '( cd whatever;zgv )'.
+ */
+
+if(stat(filename,&sbuf)!=-1 && S_ISDIR(sbuf.st_mode))
+  {
+  chdir(filename);
+  return(0);
+  }
   
+openstdin_nonblocking();	/* we hadn't done this yet */
 one_file_only=1;
 if(cfg.onefile_progress)
   {
@@ -155,8 +204,9 @@ else
   fprintf(stderr,"Loading...");
   }
   
-if(readpicture(filename,hf,1)!=0)
+if(readpicture(filename,hf,1)!=_PIC_OK)
   {
+  screenoff();
   fprintf(stderr,"\rError loading file.\n");
   exit(1);
   }
@@ -236,6 +286,20 @@ while(!quit)
       oldent=curent;
       redrawall(curent,startfrom);
       break;
+    case RK_HOME: case 'A'-0x40: /* home or ^a */
+      curent=1;
+      break;
+    case RK_END:  case 'E'-0x40: /* end  or ^e */
+      curent=gifdirsiz;
+      break;
+    case RK_PAGE_UP: case 'U'-0x40: /* pgup or ^u */
+      curent-=YSIZ*3;
+      if(curent<1) curent=1;
+      break;
+    case RK_PAGE_DOWN: case 'V'-0x40: /* pgdn or ^v */
+      curent+=YSIZ*3;
+      if(curent>gifdirsiz) curent=gifdirsiz;
+      break;
     case 'D': case RK_DELETE:   /* shift-D or the 'del' key */
       if(gifdir[curent].isdir) break;
       if(delete_file(gifdir[curent].name))
@@ -275,6 +339,22 @@ while(!quit)
         showgifdir(1,0); curent=oldent=startfrom=1;
         showbar(curent,1,startfrom);
         }
+      else if(cfg.cmd[0])
+        {
+        /* alternative command */
+        /* avoid using sprintf() just in case of stray %'s etc. */
+        char *ptr,sav;
+
+        ptr=strstr(cfg.cmd,"%s");	/* we know it's there - see rcfile.c */
+        sav=*ptr; *ptr=0;
+        strcpy(buf,cfg.cmd);
+        strcat(buf,gifdir[curent].name);
+        strcat(buf,ptr+2);
+        *ptr=sav;
+        screenoff();
+        system(buf);
+        redrawall(curent,startfrom);
+        }
       else
         {
         inithowfar();
@@ -368,7 +448,7 @@ int sofar,total;
 int f,d;
 
 /* we jump back to an abort message if Esc was pressed */
-if(!one_file_only)
+if(!one_file_only && !cfg.repeat_timer)
   if(readnbkey(zgv_ttyfd)==RK_ESC)
     {
     /* these routines blast the malloc'ed stuff, which has *for sure*
@@ -594,6 +674,7 @@ dirfile=opendir(".");
 entnum=0;
 while((anentry=readdir(dirfile))!=NULL)
   {
+  if(entnum==MDIRSIZ) break;
   if(anentry->d_name[0]=='.' && anentry->d_name[1]!='.')
     continue;	/* skip 'dot' files */
   if((strcmp(anentry->d_name,".")!=0)&&       /* no . and no .. if at root */
@@ -601,7 +682,7 @@ while((anentry=readdir(dirfile))!=NULL)
     {
     if((stat(anentry->d_name,&buf))==-1)
       buf.st_mode=0;
-    if((buf.st_mode&0x7000)==0x4000) isdir=1; else isdir=0;
+    isdir=S_ISDIR(buf.st_mode);
     /* directories, GIF/JPG/PBM/PGM/PPM tested here */
     if(((l=strlen(anentry->d_name))>4) || isdir)
       if((!strcasecmp(anentry->d_name+l-4,".gif")) ||
@@ -654,7 +735,7 @@ screenon()
 int r,g,b,n;
 unsigned char *tmp;
 
-if(cfg.xvpic_index && vga_hasmode(G640x480x256))
+if(cfg.xvpic_index && vga_hasmode(G640x480x256) && cfg.force16fs==0)
   {
   vga_setmode(G640x480x256);
   /* construct 3:3:2 palette */
@@ -678,11 +759,12 @@ else
   /* even for the normal file selector we use an 8-bit mode if possible,
    * since it's much faster.
    */
-  vga_setmode(vga_hasmode(G640x480x256)?G640x480x256:G640x480x16);
+  vga_setmode((vga_hasmode(G640x480x256) && cfg.force16fs==0)?
+                G640x480x256:G640x480x16);
   cfg.xvpic_index=0;
   barheight=GDFSIZ*6+6;
   idx_medium=MIDGREY; idx_dark=DARK; idx_light=LIGHT; idx_black=BLACK;
-  if(one_file_only && vga_hasmode(G640x480x256))
+  if(one_file_only && vga_hasmode(G640x480x256) && cfg.force16fs==0)
     idx_medium=255;	/* well, not 0 at any rate */
   }
 
@@ -755,7 +837,7 @@ for(y=0;y<logoh;y++)
 cleartext()
 {
 if(cfg.cleartext)
-  fprintf(stderr,"\x1b[H\x1b[J");
+  fprintf(stderr,"\033[H\033[J");
 }
 
 
@@ -786,7 +868,7 @@ switch(errnumber)
     strcat(buf,"corrupt or short file"); break;
   case _PICERR_SHOWN_ALREADY:
     /* this only happens with JPEGs */
-    return;
+    return(0);
   default:
     strcat(buf,"unknown error (ulp!)");
   }
@@ -862,12 +944,12 @@ if((done%5)==0 || sofar==total)
 clear_xvpic(xpos,ypos)
 int xpos,ypos;
 {
-unsigned char tmp[80];
+unsigned char tmp[86];
 int y;
 
-memset(tmp,idx_medium,80);
-for(y=0;y<60;y++)
-  vga_drawscansegment(tmp,xpos+(BARWIDTH-80)/2,ypos+y+GDFSIZ*6+9,80);
+memset(tmp,idx_medium,86);
+for(y=-3;y<63;y++)
+  vga_drawscansegment(tmp,xpos+(BARWIDTH-80)/2-3,ypos+y+GDFSIZ*6+9,86);
 }
 
 
@@ -1033,3 +1115,92 @@ for(f=1;f<=gifdirsiz;f++)
 
 usleep(400000);	/* for effect :) */
 }
+
+
+
+/* if we're not running directly on a console, try to find a free console
+ * and move us to it. Notes old VT so we can switch back to it when finished,
+ * if we ran on a different one to start off with.
+ */
+int fixvt()
+{
+struct stat sbuf;
+struct vt_stat vts;
+int major,minor;
+int fd;
+int num;
+char vt_filename[20];
+
+/* see if terminal is a console */
+fd=dup(fileno(stderr));
+fstat(fd,&sbuf);
+major=sbuf.st_rdev>>8;
+minor=sbuf.st_rdev&0xff;
+close(fd);
+if(major==4 && minor<64)
+  return(1);	/* if on a console, already ok */
+
+/* otherwise we need to look for a free VT, redirect std{in,out,err},
+ * and switch to it. If there's no free VTs, give up now.
+ */
+
+separate_vt=1;
+
+/* still root perms, so this shouldn't be a problem... */
+if((fd=open("/dev/console",O_RDONLY))<0) return(0);
+ioctl(fd,VT_GETSTATE,&vts);
+original_vt=vts.v_active;
+ioctl(fd,VT_OPENQRY,&num);
+if(num==-1) return(0);	/* no VTs free */
+
+/* now, before we go for it, we test the *current* VT to see if RW access
+ * is allowed. If so, the user's probably 'genuine'.
+ */
+sprintf(vt_filename,"/dev/tty%d",original_vt);
+stat(vt_filename,&sbuf);
+if(getuid()!=sbuf.st_uid)
+  {
+  fprintf(stderr,"You must be the owner of the current console to run zgv.\n");
+  exit(1);
+  }
+
+sprintf(vt_filename,"/dev/tty%d",num);
+setsid();
+if(cfg.dontchangestdin==0)
+  if(freopen(vt_filename,"r",stdin)==NULL)	return(0);
+if(freopen(vt_filename,"w",stderr)==NULL)	return(0);
+/* we keep stdout - doesn't need to be changed */
+
+/* switch to the new VT */
+ioctl(fd,VT_ACTIVATE,num);
+close(fd);
+/* ok, done it. */
+return(1);
+}
+
+
+void switchback()
+{
+struct vt_mode vtm;
+
+if(separate_vt)
+  {
+  fprintf(stderr,"%c[H%c[J",27,27);	/* seems to get junk-filled... */
+  ioctl(zgv_ttyfd,VT_GETMODE,&vtm);
+  vtm.mode=VT_AUTO;
+  ioctl(zgv_ttyfd,VT_SETMODE,&vtm);
+  ioctl(zgv_ttyfd,VT_ACTIVATE,original_vt);
+  }
+}
+
+
+/* prints a message saying which the old VT was if reqd. */
+oldvttell()
+{
+if(separate_vt)		/* only if needed to switch VTs */
+  fprintf(stderr,"Old VT was %d - press %sAlt-F%d to return.\n",original_vt,
+			(original_vt>12)?"Shift-":"",((original_vt-1)%12)+1);
+/* is the above message right for VTs>12? AltGr rings a bell... (well, that's
+ * ^G I s'pose but you know what I mean (groan))
+ */
+}
